package sun.security.util;

import java.lang.ref.SoftReference;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.lang.ref.ReferenceQueue;
import java.util.Map;

class MemoryCache<K, V> extends Cache<K, V>
{
    private static final float LOAD_FACTOR = 0.75f;
    private static final boolean DEBUG = false;
    private final Map<K, MemoryCache.CacheEntry<K, V>> cacheMap;
    private int maxSize;
    private long lifetime;
    private final ReferenceQueue<V> queue;
    
    public MemoryCache(boolean b, int n) {
        this(b, n, 0);
    }
    
    public MemoryCache(boolean b, int maxSize, int n) {
        this.maxSize = maxSize;
        this.lifetime = (long)(n * 1000);
        if (b) {
            this.queue = new ReferenceQueue();
        }
        else {
            this.queue = null;
        }
        this.cacheMap = new LinkedHashMap((int)((float)maxSize / 0.75f) + 1, 0.75f, true);
    }
    
    private void emptyQueue() {
        if (this.queue == null) {
            return;
        }
        this.cacheMap.size();
        while (true) {
            MemoryCache.CacheEntry<Object, Object> cacheEntry = (MemoryCache.CacheEntry)this.queue.poll();
            if (cacheEntry == null) {
                break;
            }
            Object key = cacheEntry.getKey();
            if (key == null) {
                continue;
            }
            MemoryCache.CacheEntry<Object, Object> cacheEntry2 = (MemoryCache.CacheEntry)this.cacheMap.remove(key);
            if (cacheEntry2 != null && cacheEntry != cacheEntry2) {
                this.cacheMap.put(key, cacheEntry2);
            }
            else {
                continue;
            }
        }
        return;
    }
    
    private void expungeExpiredEntries() {
        this.emptyQueue();
        if (this.lifetime == 0L) {
            return;
        }
        int n = 0;
        long currentTimeMillis = System.currentTimeMillis();
        Iterator<MemoryCache.CacheEntry<K, V>> iterator = this.cacheMap.values().iterator();
        while (iterator.hasNext()) {
            if (!((MemoryCache.CacheEntry)iterator.next()).isValid(currentTimeMillis)) {
                iterator.remove();
                ++n;
            }
            else {
                continue;
            }
        }
        return;
    }
    
    @Override
    public synchronized int size() {
        this.expungeExpiredEntries();
        return this.cacheMap.size();
    }
    
    @Override
    public synchronized void clear() {
        if (this.queue != null) {
            Iterator<MemoryCache.CacheEntry<K, V>> iterator = this.cacheMap.values().iterator();
            while (iterator.hasNext()) {
                ((MemoryCache.CacheEntry)iterator.next()).invalidate();
            }
            while (this.queue.poll() != null) {}
        }
        this.cacheMap.clear();
    }
    
    @Override
    public synchronized void put(K k, V v) {
        this.emptyQueue();
        MemoryCache.CacheEntry<K, V> cacheEntry = (MemoryCache.CacheEntry<K, V>)this.cacheMap.put(k, this.newEntry(k, v, (this.lifetime == 0L) ? 0L : (System.currentTimeMillis() + this.lifetime), this.queue));
        if (cacheEntry != null) {
            cacheEntry.invalidate();
            return;
        }
        if (this.maxSize > 0 && this.cacheMap.size() > this.maxSize) {
            this.expungeExpiredEntries();
            if (this.cacheMap.size() > this.maxSize) {
                Iterator<MemoryCache.CacheEntry<K, V>> iterator = this.cacheMap.values().iterator();
                MemoryCache.CacheEntry<Object, Object> cacheEntry2 = (MemoryCache.CacheEntry)iterator.next();
                iterator.remove();
                cacheEntry2.invalidate();
            }
        }
        return;
    }
    
    @Override
    public synchronized V get(Object o) {
        this.emptyQueue();
        MemoryCache.CacheEntry<Object, Object> cacheEntry = (MemoryCache.CacheEntry)this.cacheMap.get(o);
        if (cacheEntry == null) {
            return null;
        }
        if (!cacheEntry.isValid((this.lifetime == 0L) ? 0L : System.currentTimeMillis())) {
            this.cacheMap.remove(o);
            return null;
        }
        return cacheEntry.getValue();
    }
    
    @Override
    public synchronized void remove(Object o) {
        this.emptyQueue();
        MemoryCache.CacheEntry<Object, Object> cacheEntry = (MemoryCache.CacheEntry)this.cacheMap.remove(o);
        if (cacheEntry != null) {
            cacheEntry.invalidate();
        }
    }
    
    @Override
    public synchronized void setCapacity(int n) {
        this.expungeExpiredEntries();
        if (n > 0 && this.cacheMap.size() > n) {
            Iterator<MemoryCache.CacheEntry<K, V>> iterator = this.cacheMap.values().iterator();
            for (int i = this.cacheMap.size() - n; i > 0; --i) {
                MemoryCache.CacheEntry<Object, Object> cacheEntry = (MemoryCache.CacheEntry)iterator.next();
                iterator.remove();
                cacheEntry.invalidate();
            }
        }
        this.maxSize = ((n > 0) ? n : 0);
    }
    
    @Override
    public synchronized void setTimeout(int n) {
        this.emptyQueue();
        this.lifetime = ((n > 0) ? ((long)n * 1000L) : 0L);
    }
    
    @Override
    public synchronized void accept(Cache.CacheVisitor<K, V> cacheVisitor) {
        this.expungeExpiredEntries();
        cacheVisitor.visit(this.getCachedEntries());
    }
    
    private Map<K, V> getCachedEntries() {
        HashMap hashMap = new HashMap(this.cacheMap.size());
        for (MemoryCache.CacheEntry cacheEntry : this.cacheMap.values()) {
            hashMap.put(cacheEntry.getKey(), cacheEntry.getValue());
        }
        return hashMap;
    }
    
    protected MemoryCache.CacheEntry<K, V> newEntry(K k, V v, long n, ReferenceQueue<V> referenceQueue) {
        if (referenceQueue != null) {
            return new MemoryCache.SoftCacheEntry(k, v, n, referenceQueue);
        }
        return new MemoryCache.HardCacheEntry(k, v, n);
    }
    
    private static class HardCacheEntry<K, V> implements MemoryCache.CacheEntry<K, V>
    {
        private K key;
        private V value;
        private long expirationTime;
        
        HardCacheEntry(K key, V value, long expirationTime) {
            this.key = key;
            this.value = value;
            this.expirationTime = expirationTime;
        }
        
        @Override
        public K getKey() {
            return this.key;
        }
        
        @Override
        public V getValue() {
            return this.value;
        }
        
        @Override
        public boolean isValid(long n) {
            int n2 = n <= this.expirationTime;
            if (n2 == 0) {
                this.invalidate();
            }
            return n2;
        }
        
        @Override
        public void invalidate() {
            this.key = null;
            this.value = null;
            this.expirationTime = -1L;
        }
    }
    
    private static class SoftCacheEntry<K, V> extends SoftReference<V> implements MemoryCache.CacheEntry<K, V>
    {
        private K key;
        private long expirationTime;
        
        SoftCacheEntry(K key, V referent, long expirationTime, ReferenceQueue<V> q) {
            super(referent, q);
            this.key = key;
            this.expirationTime = expirationTime;
        }
        
        @Override
        public K getKey() {
            return this.key;
        }
        
        @Override
        public V getValue() {
            return this.get();
        }
        
        @Override
        public boolean isValid(long n) {
            int n2 = n <= this.expirationTime && this.get() != null;
            if (n2 == 0) {
                this.invalidate();
            }
            return n2;
        }
        
        @Override
        public void invalidate() {
            this.clear();
            this.key = null;
            this.expirationTime = -1L;
        }
    }
    
    private interface CacheEntry<K, V>
    {
        public abstract boolean isValid(long p0);
        
        public abstract void invalidate();
        
        public abstract K getKey();
        
        public abstract V getValue();
    }
}
